import os
import hashlib
import uuid
from typing import List
import logging

from libreary.adapter_manager import AdapterManager
from libreary.exceptions import ChecksumMismatchException
from libreary.adapters import LocalAdapter
from libreary.metadata import SQLite3MetadataManager
from libreary.exceptions import NoCopyExistsException

logger = logging.getLogger(__name__)


class Ingester:

    def __init__(self, config: dict, metadata_man: object = None):
        """
        Constructor for the Ingester object. This object can be created manually, but
        in most cases, it will be constructed by the LIBRE-ary main object. It expects a python dict
        :param config, which should be structured as follows:
        ```{json}
        {
                "metadata": {
                    "db_file": "path to SQLite3 DB file for metadata"
                },
                "adapters": # List of adapters - each entry should look like:
                [{
                    "type":"AdapterType (name of class)",
                    "id": "Adapter Identifier"
                }],
                "options": {
                    "dropbox_dir": "Path to dropbox directory, where files you want to ingest should be placed",
                    "output_dir": "Path to directory you want files to be retrieved to",
                    "config_dir": "Path to config directory"
                },
                "canonical_adapter":"Adapter Identifier for Canonical Adapter"
            }
        ```
        """
        try:
            self.config = config
            self.dropbox_dir = config["options"]["dropbox_dir"]
            self.canonical_adapter_id = config["canonical_adapter"]
            self.canonical_adapter_type = config["canonical_adapter_type"]
            self.config_dir = config["options"]["config_dir"]

            self.metadata_man = metadata_man
            if self.metadata_man is None:
                raise KeyError

            logger.debug("Ingester configuration valid, creating Ingester.")
        except KeyError as e:
            logger.error("Ingester Configuration Invalid")
            raise e

    def ingest(self, current_file_path: str, levels: List[str],
               description: str, delete_after_store: bool = False, metadata_schema: List = [], metadata: List = []) -> str:
        """
        Ingest an object to LIBREary. This method:
        - Creates the canonical copy of the object
        - Creates the entry in the `resources` table describing the resource
        - Optionally, delete the file out of the dropbox dir.

        :param current_file_path -
        """
        filename = current_file_path.split("/")[-1]
        sha1Hash = hashlib.sha1(open(current_file_path, "rb").read())
        checksum = sha1Hash.hexdigest()

        canonical_adapter = AdapterManager.create_adapter(
            self.canonical_adapter_type, self.canonical_adapter_id, self.config_dir, self.config["metadata"])

        obj_uuid = str(uuid.uuid4())

        logger.debug(f"Ingesting resource {obj_uuid} with filename {filename}")

        canonical_adapter_locator = canonical_adapter._store_canonical(
            current_file_path, obj_uuid, checksum, filename)

        levels = ",".join([str(l) for l in levels])

        # Ingest to db
        self.metadata_man.ingest_to_db(
            canonical_adapter_locator,
            levels,
            filename,
            checksum,
            obj_uuid,
            description)

        # Ingest file metadata:
        if len(metadata_schema) == len(metadata) and len(metadata_schema) != 0:
            self.metadata_man.set_object_metadata_schema(
                obj_uuid, metadata_schema)
            self.metadata_man.set_all_object_metadata(obj_uuid, metadata)

        # If file is not in dropbox, copy it there

        if delete_after_store:
            os.remove(current_file_path)

        return obj_uuid

    def verify_ingestion(self, r_id: str) -> bool:
        """
        Make sure an object has been properly ingested.

        :param r_id - the UUID of the resource you are verifying
        """
        pass

    def list_resources(self) -> List[List[str]]:
        """
        Return a list of summaries of each resource. This summary includes:

        `id`, `path`, `levels`, `file name`, `checksum`, `object uuid`, `description`

        This method trusts the metadata database. There should be a separate method to
        verify the metadata db so that we know we can trust this info
        """
        return self.metadata_man.list_resources()

    def delete_resource(self, r_id: str) -> None:
        """
        Delete a resource from the LIBREary.

        This method deletes the canonical copy and removes the corresponding entry in the `resources`
            table.

        :param r_id - the UUID of the resouce you're deleting
        """

        try:
            resource_info = self.metadata_man.get_resource_info(r_id)[0]
            canonical_checksum = resource_info[4]
        except IndexError:
            logger.debug(f"Already deleted {r_id}")

        canonical_adapter = AdapterManager.create_adapter(
            self.canonical_adapter_type, self.canonical_adapter_id, self.config_dir, self.config["metadata"])

        try:
            checksum = canonical_adapter.get_actual_checksum(r_id)
        except NoCopyExistsException:
            self.metadata_man.delete_resource(r_id)
            return

        if checksum == canonical_checksum:
            logger.debug(f"Deleting canonical copy of object {r_id}")
            canonical_adapter._delete_canonical(r_id)
        else:
            raise ChecksumMismatchException

        logger.debug(f"Deleting object {r_id} from resources database")
        self.metadata_man.delete_resource(r_id)



import json
import os
import string
import random
import hashlib
import shutil
from typing import List
import logging
import ast

from libreary.adapters.BaseAdapter import BaseAdapter
from libreary.adapters.local import LocalAdapter
from libreary.adapters.s3 import S3Adapter
from libreary.adapters.drive import GoogleDriveAdapter
from libreary.exceptions import ResourceNotIngestedException, ChecksumMismatchException, NoCopyExistsException
from libreary.exceptions import RestorationFailedException, AdapterCreationFailedException, AdapterRestored
from libreary.metadata import SQLite3MetadataManager

logger = logging.getLogger(__name__)

adapters_translate_table = {
    "LocalAdapter": LocalAdapter,
    "S3Adapter": S3Adapter,
    "GoogleDriveAdapter": GoogleDriveAdapter,
}
metadata_man_translate_table = {
    "SQLite3MetadataManager": SQLite3MetadataManager,
}


class AdapterManager:
    """
    The AdapterManager is responsible for all interaction with adapters, except for intial ingestion.

    It is able to keep track of all of the adapters we have, do integrity checks on them,
    perform initial distribution, compare versions from different adapters, and make insert and delete calls.

    The Adapter Manager is responsible for most of the operation of ingestion, deletion, and management of
    digital objects within LIBREary. Most customization will occur by subclassing the AdapterManager.

    This class currently contains the following methods:

    - reload_levels_adapters (create adapter objects and set levels based on configuration)
    - get_all_levels (get levels based on what exists in metadata db)
    - get_all_adapters (create adapter objects based on levels)
    - set_additional_adapter (manually create an adapter object and add it to the AdapterManager's list of adapters)
    - verify_adapter (make sure an adapter is working properly)
    - create_adapter [static method] (factory function for adapter objects)
    - send_resource_to_adapters (send copies a resource to all the places they need to be)
    - get_adapters_by_level (get all adapters from a level)
    - delete_resource_from_adapters (delete non-canonical copies of an object)
    - change_resource_level (change the level of an object)
    - get_canonical_copy_metadata
    - summarize_copies
    - retrieve_by_preference (retrieve an object, prefering the canonical adapter)
    - check_single_resource_single_adapter (make sure that a copy matches the canonical copy)
    - verify_adapter_metadata (verify that a file can be retrieved)
    - get_resource_metadata
    - restore_canonical_copy (restore a faulty canonical copy)
    - compare_copies (check if two copies of a resource are the same)
    - verify_copy (check if a copy matches the canonicl copy)
    """

    def __init__(self, config: dict, metadata_man: object = None):
        """
        Constructor for the AdapterManager object. This object can be created manually, but
        in most cases, it will be constructed by the LIBRE-ary main object. It expects a python dict
        :param config, which should be structured as follows:
        ```{json}
        {
                "metadata": {
                    "db_file": "path to SQLite3 DB file for metadata"
                },
                "adapters": # List of adapters - each entry should look like:
                [{
                    "type":"AdapterType (name of class)",
                    "id": "Adapter Identifier"
                }],
                "options": {
                    "dropbox_dir": "Path to dropbox directory, where files you want to ingest should be placed",
                    "output_dir": "Path to directory you want files to be retrieved to",
                    "config_dir": "Path to config directory"
                },
                "canonical_adapter":"Adapter Identifier for Canonical Adapter"
            }
        ```
        """
        try:
            self.config = config
            self.all_adapters = self.config["adapters"]
            self.dropbox_dir = config["options"]["dropbox_dir"]
            self.ret_dir = config["options"]["output_dir"]
            self.config_dir = config["options"]["config_dir"]
            self.levels = {}
            self.adapters = {}
            self.canonical_adapter = self.config["canonical_adapter"]
            self.metadata_man = metadata_man
            if self.metadata_man is None:
                raise KeyError
            logger.debug(
                "Adapter Manager Configuration Valid, creating Adapter Manager")
        except KeyError:
            logger.error("Adapter Manager Configuration Invalid")
            raise KeyError
        # Run this any time you expect levels and adapters to change
        # For most use cases, this will only be on construction
        # This method should be run externally, any time a new level is added
        self.reload_levels_adapters()

    def reload_levels_adapters(self) -> None:
        """
        Set the `self.adapters` and `self.levels` instance variables.

        This object needs to be stateful in this way because each adapter might
        either require time-sensitive authentication information (tokens, etc), or
        may be computationally expensive to create. For this reason, we want the
        `AdapterManager` to have instance variables with adapter objects.
        """
        self._set_levels()
        self._set_adapters()

    def get_all_levels(self) -> List[dict]:
        """
        Returns all levels in the metadata database.

        Returns a list of dictionaries each with the following format:
        ```
        {"id": (int) level ID,
         "name": (str) level name,
         "frequency": (int) scheduled check frequency,
         "adapters": (dict) dictionary of adapters associated with this level}
        ```
        """
        level_data = self.metadata_man.get_levels()
        levels = {}
        for level in level_data:
            levels[level[1]] = {"id": level[0],
                                "name": level[1],
                                "frequency": level[2],
                                "adapters": json.loads(level[3])}
            logger.debug(
                f"Found level {level[1]} with adapters {json.loads(level[3])}")
        return levels

    def _set_levels(self) -> None:
        """
        Convenience method to reset levels as an instance variable
        """
        self.levels = self.get_all_levels()

    def get_all_adapters(self) -> List[dict]:
        """
        Set up all of the adapters we will need, based on all levels
        from the metadata database.

        Parses `self.levels` to create adapter objects.

        The structure of the return value is is a dictionary structured as follows:

        ```
        {
            "adapter_id1": <AdapterObject>,
            "adapter_id2": <AdapterObject>,...
        }
        ```

        Ensure that `self.levels` is set properly before running this, by
        calling `self._set_levels()`
        """
        adapters = {}
        for level in self.levels.values():
            # Each level may need several adapters
            for adapter in level["adapters"]:
                adapters[adapter["id"]] = AdapterManager.create_adapter(
                    adapter["type"], adapter["id"], self.config_dir, self.config["metadata"])
                logger.debug(
                    f"Created adapter {adapter['id']} of type {adapter['type']}")
        logger.debug(f"Summary of all adapters: {adapters}")
        return adapters

    def _set_adapters(self) -> None:
        """
        Convenience method to set `self.adapters` to the output of `get_all_adapters

        The structure of `self.adapters` is is a dictionary structured as follows:

        ```
        {
            "adapter_id1": <AdapterObject>,
            "adapter_id2": <AdapterObject>,...
        }
        ```

        Ensure that `self.levels` is set properly before running this, by
        calling `self._set_levels()`
        """
        self.adapters = self.get_all_adapters()

    def set_additional_adapter(self, adapter_id: str,
                               adapter_type: str) -> BaseAdapter:
        """
        Manually add an adapter to the pool of adapters.

        :param adapter_id - the adapter ID of the adapter you're creating
            There should be a matching config file for this adapter ID

        :param adapter_type - the type of the adapter you wish to create.
            Must be the actual class name, i.e. "LocalAdapter".
        """
        adapter = AdapterManager.create_adapter(
            adapter_type, adapter_id, self.config_dir, self.config["metadata"])
        self.adapters[adapter_id] = adapter
        logger.debug(
            f"Manually added adapter {adapter_id} of type {adapter_type}")
        return adapter

    def verify_adapter(self, adapter_id: str) -> bool:
        """
        Make sure an adapter is working. To do this, we store, retrieve,
        and delete a file that we know the contents of,
        and make sure the checksums are as they should be.

        :param adapter_id - The adapter ID you'd like to verify.
        """
        dropbox_path = "{}/libreary_test_file.txt".format(
            self.config["options"]["dropbox_dir"])
        adapter = self.adapters[adapter_id]
        data_to_store = ''.join(random.choice(string.ascii_letters)
                                for i in range(500))
        with open(dropbox_path, "w") as fh:
            fh.write(data_to_store)
        real_checksum = hashlib.sha1(
            open(dropbox_path, "rb").read()).hexdigest()
        r_id = "LIBREARY_TEST_RESORURCE"

        # To circumvent full ingestion process, we manually use _ingest_canonical
        # Not recommended for end users to do this.
        locator = adapter._store_canonical(
            dropbox_path,
            r_id,
            real_checksum,
            "libreary_test_resource.txt")

        self.metadata_man.minimal_test_ingest(locator, real_checksum, r_id)

        new_path = adapter.retrieve(r_id)
        new_checksum = hashlib.sha1(open(new_path, "rb").read()).hexdigest()

        r_val = False
        if new_checksum == real_checksum:
            logger.debug(f"Verified Adapter {adapter_id}")
            r_val = True

        adapter._delete_canonical(r_id)

        try:
            os.remove(new_path)
        except Exception:
            return

        os.remove(dropbox_path)

        # Delete from res table
        self.metadata_man.delete_resource(r_id)

        return r_val

    @staticmethod
    def create_adapter(adapter_type: str, adapter_id: str,
                       config_dir: str, metadata_man_config: dict, metadata_man_type="SQLite3MetadataManager") -> BaseAdapter:
        """
        Static method for creating and returning an adapter object.
        This is essentially an Adapter factory.

        :param adapter_type - must be the name of a valid adapter class.
        :param adapter_id - the identifier you want to label this adapter with
        :param config_dir - configuration directory. Must contain a file called
            `{adapter_id}_config.json`
        """
        cfg = AdapterManager.create_config_for_adapter(
            adapter_id, adapter_type, config_dir)
        adapter = adapters_translate_table[adapter_type](
            ast.literal_eval(f"{cfg}"),
            metadata_man_translate_table[metadata_man_type](
                ast.literal_eval(f"{metadata_man_config}")))

        return adapter

    @staticmethod
    def create_config_for_adapter(
            adapter_id: str, adapter_type: str, config_dir: str) -> dict:
        """
        Static method for creating an adapter configuration. This is necessary for
        the adapter factory.

        :param adapter_type - must be the name of a valid adapter class.
        :param adapter_id - the identifier you want to label this adapter with
        :param config_dir - configuration directory. Must contain a file called
            `{adapter_id}_config.json`
        """
        base_config = json.load(
            open("{}/{}_config.json".format(config_dir, adapter_id)))
        general_config = json.load(
            open("{}/config.json".format(config_dir)))

        full_adapter_conf = {}
        full_adapter_conf["adapter"] = base_config["adapter"]
        full_adapter_conf["adapter"]["adapter_type"] = adapter_type
        full_adapter_conf["metadata"] = general_config["metadata"]
        full_adapter_conf["options"] = general_config["options"]

        return full_adapter_conf

    def send_resource_to_adapters(
            self, r_id: str, delete_after_send: bool = False) -> None:
        """
        Sends a resource to all the places it should go. The resource must
        have already been ingested through the Ingester. This method:
            1. Figures out what levels a resource has been assigned
            2. Figures out what adapters are associated with that level
            3. Figured out any overlap, to avoid storing things twice in one adapter
            4. Stores copies to each adapter
            5. optionally, deletes any remaining files in the dropbox directory

            :param r_id - resource UUID you wish to distribute
            :param delete_after_send - boolean indicating whether to delete
                files after storage
        """
        try:
            resource_metadata = self.get_resource_metadata(r_id)[0]
        except IndexError:
            raise ResourceNotIngestedException

        # Make sure that resource is in dropbox:
        filename = resource_metadata[3]
        expected_location = "{}/{}".format(self.dropbox_dir, filename)

        file_there = False
        if os.path.isfile(expected_location):
            file_hash = hashlib.sha1(
                open(
                    expected_location,
                    "rb").read()).hexdigest()
            expected_hash = resource_metadata[4]
            if file_hash == expected_hash:
                # there's a file in that location, and its checksum matches
                file_there = True

        # If the file isn't where we want it, put it there
        if not file_there:
            logger.debug(
                f"Could not find object {r_id} in Dropbox Directory. Moving it")
            # retrieve moves it to the retrieval dir
            current_path = self.adapters[self.canonical_adapter].retrieve(r_id)
            # so we move it to the dropbox_dir
            shutil.move(current_path, expected_location)

        levels = resource_metadata[2].split(",")
        for level in levels:
            adapters = self.get_adapters_by_level(level)
            for adapter in adapters:
                logger.debug(f"Storing object {r_id} to adapter {adapter}")
                adapter.store(r_id)

        if delete_after_send:
            logger.debug(f"Deleting object {r_id} after send")
            os.remove(expected_location)

    def get_adapters_by_level(self, level: str) -> List[BaseAdapter]:
        """
        Get a list of adapter objects based on a level.
        Returns a list of callable adapter objects.

        :param level - the name of the level you want the adapters for
        """
        adapter_names = self.levels[level]["adapters"]
        adapters = []
        for adapter in adapter_names:
            adapters.append(self.adapters[adapter["id"]])
        return adapters

    def delete_resource_from_adapters(self, r_id: str) -> None:
        """Deletes a resource from all adapters it's stored in.
           Does not delete canonical copy

           :param r_id - UUID of resource to delete copies of
        """
        try:
            resource_metadata = self.get_resource_metadata(r_id)[0]
        except IndexError:
            raise ResourceNotIngestedException

        levels = resource_metadata[2].split(",")
        for level in levels:
            adapters = self.get_adapters_by_level(level)
            for adapter in adapters:
                logger.debug(f"Deleting object {r_id} from {adapter}")
                adapter.delete(r_id)

    def change_resource_level(self, r_id: str, new_levels: List[str]) -> None:
        """
        Assign a new set of levels to a resource.
        Removes all levels from a resource, replaces them with :param new_levels

        :param r_id - UUID of resource you'd like to change the levels of
        :param new_levels: list of names of levels to assign to the resource
        """

        logger.debug(f"Changing object {r_id} to new levels: {new_levels}")
        # Because d_r_f_a doesn't delete canonical copy, we can simply use
        # it to reset
        self.delete_resource_from_adapters(r_id)
        self.metadata_man.update_resource_levels(r_id, new_levels)
        # now that we've updated the levels, we refresh levels
        self.reload_levels_adapters()
        # now, we can just act as if it has never been sent off:
        self.send_resource_to_adapters(r_id)

    def summarize_copies(self, r_id: str) -> List[List[str]]:
        """
        Get a summary of all copies of a single resource. That summary includes:

        `copy_id`, `resource_id`, `adapter_identifier`, `locator`, `checksum`, `adapter type`, `canonical (bool)`
        for each copy

        This method trusts the metadata database. There should be a separate method to
        verify the metadata db so that we know we can trust this info

        :param r_id - UUID of resource you'd like to learn about
        """
        return self.metadata_man.summarize_copies(r_id)

    def get_canonical_copy_metadata(self, r_id: str) -> List[List[str]]:
        """
        Get a summary of the canonical copy of an object's medatada. That summary includes:
        `copy_id`, `resource_id`, `adapter_identifier`, `locator`, `checksum`, `adapter type`, `canonical (bool)`

        :param r_id - UUID of resource you'd like to learn about
        """
        return self.metadata_man.get_canonical_copy_metadata(self, r_id)

    def retrieve_by_preference(self, r_id: str) -> str:
        """
        Retrieve a resource.

        Get a copy of a file, preferring canonical adapter, then enforcing some preference hierarchy
        This will be called when Libreary is asked to retrieve.

        This places a file in the configured `output_dir` and returns a path to the retrieved file.

        Keep in mind that the output directory may be volatile and should not be used for storage.

        :param r_id - UUID of resource you'd like to retrieve
        """
        # First, try the canonical copy:
        try:
            new_loc = self.adapters[self.canonical_adapter].retrieve(r_id)
            return new_loc
        except ChecksumMismatchException:
            logger.error(
                "Canonical Recovery Failed. Attempting to Restore Canonical Copy")
            self.restore_canonical_copy(r_id)

        for adapter in self.adapters.values():
            try:
                new_loc = adapter.retrieve(r_id)
                return new_loc
            except ChecksumMismatchException:
                logger.error(
                    "Canonical Recovery Failed. Attempting to Restore Canonical Copy")
                self.restore_from_canonical_copy(adapter.adapter_id, r_id)

    def check_single_resource_single_adapter(
            self, r_id: str, adapter_type: str, adapter_id: str) -> bool:
        """
        Ensure that a copy of an object matches its canonical checksum.
        This method trusts that the metadata db has the proper canonical checksum.

        If a copy is found to be faulty, a restore is attempted.
        If a copy is fount to not exist, it is created.

        :param r_id - UUID of resource you'd like to check
        :param r_id - adapter_id for copy of resource you are checking
        """
        resource_info = self.get_resource_metadata(r_id)
        canonical_checksum = resource_info[4]

        copies = self.get_all_copies_metadata(r_id)

        for copy in copies:
            if adapter_id == copy[2]:
                found = True
                if copy[2] != canonical_checksum:
                    try:
                        logger.debug(
                            f"Trying to restore resource {r_id} from canonical copy")
                        self.restore_from_canonical_copy(
                            adapter_id, r_id)
                    except RestorationFailedException:
                        logger.error(
                            "Restoration of {} in {} failed".format(
                                r_id, adapter_id))
                        found = False
            # didn't find the copy from this adapter
        if not found:
            logger.debug(
                f"Could not find resource {r_id} in adapter {adapter_id}")
            try:
                a = AdapterManager.create_adapter(
                    adapter_type, adapter_id, self.config_dir, self.config["metadata"])
                a.store(r_id)
                found = True
            except AdapterCreationFailedException:
                logger.error("Could not create adapter {adapter_id}")
                found = False
        return found

    def verify_adapter_metadata(
            self, adapter_id: str, r_id: str, delete_after_check: bool = True) -> bool:
        """
        Ensure that a copy of an object matches its canonical checksum.
        This method does not trust that the metadata db has the proper
        canonical checksum.

        Verifies that the file is actually retirevable via
        adapter id, not just there according to the metadata.

        Note, this retrieves the file, so it's relatively expensive.

        If a copy is found to be faulty, a restore is attempted.
        If a copy is fount to not exist, it is created.

        :param r_id - UUID of resource you'd like to check
        :param r_id - adapter_id for copy of resource you are checking
        """
        current_resource_info = self.get_resource_metadata(r_id)
        recorded_checksum = current_resource_info[4]
        current_path = self.adapters[adapter_id].retrieve(r_id)
        sha1Hash = hashlib.sha1(open(current_path, "rb").read())
        new_checksum = sha1Hash.hexdigest()

        r_val = True

        if new_checksum != recorded_checksum:
            try:
                logger.debug(f"Restoring copy of {r_id} in {adapter_id}")
                self.restore_from_canonical_copy(adapter_id, r_id)
            except RestorationFailedException:
                logger.error(
                    "Restoration of {} in {} failed".format(
                        r_id, adapter_id))
                r_val = False

        if delete_after_check:
            os.remove(current_path)

        return r_val

    def get_resource_metadata(self, r_id: str) -> List[List[str]]:
        """
        Get a summary of information about a resource. That summary includes:

        `id`, `path`, `levels`, `file name`, `checksum`, `object uuid`, `description`

        This method trusts the metadata database. There should be a separate method to
        verify the metadata db so that we know we can trust this info

        :param r_id - UUID of resource you'd like to learn about
        """
        return self.metadata_man.get_resource_info(r_id)

    def restore_canonical_copy(self, r_id: str) -> None:
        """
        Attempt to Restore a detected fault in the canonical copy of an object.

        Delete the canonical copy of an object, but keep non-canonical copies.
        After that, create a new canonical copy, preserving resource UUID,
            but with the correct object contents.

        :param r_id - UUID of resource you'd like to restore
        """
        try:
            resource_info = self.get_resource_metadata(r_id)
            real_checksum = resource_info[4]
            levels = resource_info[2].split(",")
            filename = resource_info[3]
        except IndexError:
            raise ResourceNotIngestedException

        self.adapters[self.canonical_adapter]._delete_canonical(r_id)

        current_location = 0

        try:
            for level in levels:
                adapters = self.get_adapters_by_level(level)
                for adapter in adapters:
                    try:
                        logger.debug(
                            f"Trying to restore copy of {r_id} from adapter {adapter}")
                        current_location = adapter.retrieve(r_id)
                        raise AdapterRestored
                    except ResourceNotIngestedException:
                        continue
                    except ChecksumMismatchException:
                        continue
                    except NoCopyExistsException:
                        continue
                logger.error(f"Failed to restore copy of {r_id}")
                return
        except AdapterRestored:
            self.adapters[self.canonical_adapter].store_canonical(
                current_location, r_id, real_checksum, filename)

    def restore_from_canonical_copy(self, adapter_id: str, r_id: str) -> None:
        """
        Restore a copy of an object from its canonical copy.
        To restore from the canonical copy, we can simply delete and
            re-ingest the fraudulent copy.

        :param adapter_id - the ID of the adapter with the broken copy
        :param r_id - The resource UUID of the resource we've detected an issue with
        """
        logger.debug(
            f"Restoring object {r_id} in adapter {adapter_id} from canonical copy.")
        self.adapters[adapter_id].delete(r_id)
        self.adapters[adapter_id].store(r_id)

    def compare_copies(self, r_id: str, adapter_id_1: str,
                       adapter_id_2: str, deep: bool = False) -> bool:
        """
        Compare copies of a resource in two adapters. Returns True iff
            the checksums of each copy match.

        A deep compare will actually compute the current checksum of the
            file stored in the adapter specified. Some adapters can do this
            with no file I/O, while others will have to actually retrieve the file
            to perform this operation

        :param r_id - the UUID of the resource to compare
        :param adapter_id_1 - Adapter ID of the first adapter
        :param adapter_id_2 - Adapter ID of the second adapter
        :param deep - specify whether to run a deep or shallow check
        """
        try:
            copy_info_1 = self.metadata_man.get_copy_info(r_id, adapter_id_1)
            copy_info_2 = self.metadata_man.get_copy_info(r_id, adapter_id_2)
        except IndexError:
            logger.error(f"No copy of object {r_id} exists.")
            raise NoCopyExistsException

        if not deep:
            return copy_info_1[4] == copy_info_2[4]

        return self.adapters[adapter_id_1].get_actual_checksum(
            r_id) == self.adapters[adapter_id_2].get_actual_checksum(r_id)

    def verify_copy(self, r_id: str, adapter_id: str,
                    deep: bool = False) -> bool:
        """
        Compare copies of a resource in two adapters, one being canonical.
        Returns True iff the checksums of each copy match.

        A deep compare will actually compute the current checksum of the
            file stored in the adapter specified. Some adapters can do this
            with no file I/O, while others will have to actually retrieve the file
            to perform this operation

        :param r_id - the UUID of the resource to compare
        :param adapter_id - Adapter ID of the adapter to check against
        :param deep - specify whether to run a deep or shallow check
        """
        return self.compare_copies(
            r_id, adapter_id, self.canonical_adapter, deep=deep)



import json
from typing import List
import logging

from libreary.adapter_manager import AdapterManager
from libreary.ingester import Ingester
from libreary.metadata.sqlite3 import SQLite3MetadataManager

logger = logging.getLogger(__name__)

metadata_manager_translate_table = {
    "SQLite3MetadataManager": SQLite3MetadataManager
}


class Libreary:
    """
    This is the user-facing class for LIBRE-ary. Users of LIBRE-ary should only interact
    with this class directly. LIBRE-ary objects are able to handle all of the
    functionality of this module. Developers should feel free to extend the
    functionality of this class and are encouraged to submit pull requests
    to the main repository.

    This class currently contains the following methods:

    - ingest (load a resource into LIBRE-ary)
    - retrieve (retrieve a copy of an object)
    - delete (delete an object)
    - update (update an object)
    - search (search for information about objects)
    - run_full_check (check all resources to verify integrity)
    - check_single_resource (check only a single resource)
    """

    def __init__(self, config_dir: str):
        """
        Constructor for Libreary object.
            :param config_dir - a string pointing to a directory
            containing Libreary configuration. In the config directory
            must be a file called `config.json` which contains main
            configuration, and a separate config file for each adapter
            you plan to use. More detail on adapter configs can be found
            in the adapters constructor documentation.


            The structure of `config_dir/config.json` should be as follows:
            ```{json}
            {
                "metadata": {
                    "db_file": "path to SQLite3 DB file for metadata"
                },
                "adapters": # List of adapters - each entry should look like:
                [{
                    "type":"AdapterType (name of class)",
                    "id": "Adapter Identifier"
                }],
                "options": {
                    "dropbox_dir": "Path to dropbox directory, where files you want to ingest should be placed",
                    "output_dir": "Path to directory you want files to be retrieved to",
                    "config_dir": "Path to config directory"
                },
                "canonical_adapter":"Adapter Identifier for Canonical Adapter"
            }
            ```

            The canonical adapter is the adapter which will store the "canonical" copy
            of each resource, which will then be used as the "real" version of that digital object.

            This object creeates an adapter manager and an ingester. For more information,
            see those classes.

            The output and dropbox directories are volatile and should not be used for object storage.
        """
        # Config stuff
        self.config_dir = config_dir
        self.config_path = "{}/config.json".format(self.config_dir)
        self.config = json.load(open(self.config_path))

        try:
            # Directories we care about
            self.dropbox_dir = self.config["options"]["dropbox_dir"]
            self.ret_dir = self.config["options"]["output_dir"]

            # Objects we need
            self.metadata_man = metadata_manager_translate_table[self.config["metadata"]["manager_type"]](
                self.config["metadata"])
            self.adapter_man = AdapterManager(
                self.config, metadata_man=self.metadata_man)
            self.ingester = Ingester(
                self.config, metadata_man=self.metadata_man)
            logger.debug("LIBREary configuration valid. Proceeding.")
        except KeyError:
            logger.error("Invalid LIBREary config. Exiting.")
            raise KeyError

    def run_check(deep: bool = False) -> List[str]:
        """
        Check all of the objects in the LIBRE-ary. This follows the following process:

        For each object:
            Get canonical copy and actual checksum. Make sure canonical copy matches expected checksum
            If it doesn't:
                Attempt to recover canonical copy
            Get a list of all levels that the object has been labelled as:
            For each level:
                Get a list of adapters is is stored in:
                For each adapter:
                    Check to make sure that copy's checksum matches what it should:
                    If it doesn't:
                        Attempt to recover it.


        :param deep speficies whether to use a deep search. A deep search will calculate actual checksums
        of each copy of each object, while a shallow one will trust that the checksum in the metadata
        database matches that of the actual object.
        """
        logger.debug(f"Running check of all objects in LIBREary. Deep: {deep}")

    def ingest(self, current_file_path: str, levels: List[str],
               description: str, delete_after_store: bool = False, metadata_schema: List[str] = [], metadata: List[dict] = []) -> str:
        """
        Ingest a new object to the LIBRE-ary. This:
            1. Creates an entry in the `resources` table in the metadata db
            2. Creates an object UUID
            3. Ingests the canonical copy
            4. Sends copies to adapters which match specified levels
            5. Returns object ID

        :param current_file_path - the current path to the file you wish to ingest
        :param levels - a list of names of levels. These levels must exist in the
            `levels` table in the metadata db
        :param description - a description of this object. This is useful when you
            want to search for objects later
        :param delete_after_store - Boolean. If True, the Ingester will delete the object after it's stored.
        :param metadata_schema - list of names of metadata fields to be associated with object
        :param metadata - values of the fields in the metadata schema. Structured like this:
            [{"field": "collection", "value": "cats"},
             {"field": "owner", "value": "ben glick"}]
        """

        # Don't want ingester to delete it, because then AM will need to
        # retrieve.
        obj_id = self.ingester.ingest(
            current_file_path,
            levels,
            description,
            delete_after_store=False,
            metadata_schema=metadata_schema,
            metadata=metadata)
        self.adapter_man.send_resource_to_adapters(
            obj_id, delete_after_send=delete_after_store)
        logger.debug(
            f"Ingesting object {obj_id} to LIBREary. Description: {description}")
        return obj_id

    def retrieve(self, r_id: str) -> str:
        """
        Retrieve an object. This will save a copy of the object
            as `<self.output_dir>/<object_filename>`

        The output and dropbox directories are volatile and should not be used for object storage.

        Adapters and other objects frequently may delete or write files in these directories.

        :param r_id - The resource UUID that corresponds to the object you'd like to retrieve.

        Returns a path to the retireved object.
        """
        logger.debug(f"Retrieving object {r_id}")
        new_location = self.adapter_man.retrieve_by_preference(r_id)
        return new_location

    def delete(self, r_id: str) -> None:
        """
        Delete an object from LIBRE-ary. This:
            1. Deletes the resource from all of the adapters it was stored in
            2. Deletes the resource from the metadata db entirely
            3. Removes the canonical copy

        Be careful with this function, as there is no undo option.
        """
        logger.debug(f"Deleting object {r_id}")

        # Delete metadata only if object has metadata
        if len(self.metadata_man.list_object_metadata_schema(r_id)) > 0:
            self.metadata_man.delete_object_metadata_entirely(r_id)

        self.adapter_man.delete_resource_from_adapters(r_id)
        self.ingester.delete_resource((r_id))

    def update(self, r_id: str, updated_path: str) -> None:
        """
        Update a resource with a new object. Preserves UUID and all other metadata (levels, etc.)

        :param r_id - the UUID of the object you'd like to update
        :param updated_path - path to the contents of the updated object.

        """
        logger.debug(f"Updating object {r_id}")

    def search(self, search_term: str) -> List[str]:
        """
        Search the metadata db for information about resources.

        :param search_term - a string with which to search against the metadata db.
            Can match UUID, filename, original path, or description.
        """
        return self.metadata_man.search(search_term)

    def check_single_resource(self, r_id: str, deep: bool = False) -> bool:
        """
        Check a single object in the LIBRE-ary. This follows the following process:


        Get canonical copy and actual checksum. Make sure canonical copy matches expected checksum
        If it doesn't:
            Attempt to recover canonical copy
        Get a list of all levels that the object has been labelled as:
        For each level:
            Get a list of adapters is is stored in:
            For each adapter:
                Check to make sure that copy's checksum matches what it should:
                If it doesn't:
                    Attempt to recover it.


        :param deep speficies whether to use a deep search. A deep search will calculate actual checksums
        of each copy of each object, while a shallow one will trust that the checksum in the metadata
        database matches that of the actual object.

        :param r_id - the resource ID of the object you'd like to check
        """
        logger.debug(f"Checking object {r_id}")

    def add_level(self, name: str, frequency: int,
                  adapters: List[dict], copies=1) -> None:
        """
        Add a level to the metadata database.

        :param name - name for the level
        :param frequency - check frequency for level. Currently unimplemented
        :param adapters - dict object specifying adapters the level uses. Example:
            ```{json}
            [
                {
                "id": "local1",
                "type":"LocalAdapter"
                },
                {
                "id": "local2",
                "type":"LocalAdapter"
                }
            ]

            ```
        :param copies - copies to store for each adapter. Currently, only 1 is supported
        """
        logger.debug(f"Adding new level: {name}")
        self.metadata_man.add_level(name, frequency, adapters, copies=1)
        self.adapter_man.reload_levels_adapters()

    def delete_level(self, level_name: str):
        """
        Delete a level from the metadata database.

        :param level_name - name of the level to delete
        """
        logger.debug(f"Deleting Level {level_name}")
        self.metadata_man.remove_level(level_name)
        self.adapter_man.reload_levels_adapters()



class BaseAdapter:
    """
    Class definition for the base adapter for libre-ary resource adapters.

    Use this mostly as a reference for what other adapters need to implement.

    An adapter allows LIBREary to save copies of digital objects
        to different places across cyberspace. Working with many
        adapters in concert, one should be able do save sufficient
        copies to places they want them.
    """

    def __init__(config: dict):
        """
        This should handle configuration (auth, etc.) and set up the
        metadata db connection
        """
        pass

    def store(resource_id: str) -> str:
        """
        Store a copy of a resource in this adapter.

        Store assumes that the file is in the `dropbox_dir`.
        AdapterManager will always verify that this is the case.

        :param r_id - the resource to store's UUID
        """
        pass

    def _store_canonical(current_path: str, r_id: str) -> str:
        """
            Store a canonical copy of a resource in this adapter.

            If we're using an adapter as a canonical adapter, we need
            to be able to store from a current path, taking in a generated UUID,
            rather than looking info up from the database.

            :param current_path - current path to object
            :param r_id - UUID of resource you're storing
            :param checksum - checksum of resource
            :param filename - filename of resource you're storing

        """
        pass

    def _delete_canonical(self, r_id: str) -> None:
        """
        Delete a canonical copy of a resource from this adapter.
        Delete the corresponding entry in the `copies` table.

        :param r_id - the resource to retrieve's UUID
        """
        pass

    def retrieve(resource_id: str) -> str:
        """
        Retrieve a copy of a resource from this adapter.

        Retrieve assumes that the file can be stored to the `output_dir`.
        AdapterManager will always verify that this is the case.

        Returns the path to the resource.

        May overwrite files in the `output_dir`

        :param r_id - the resource to retrieve's UUID
        """
        pass

    def update(resource_id: str, updated: str) -> None:
        """
        Update a resource with a new object. Preserves UUID and all other metadata (levels, etc.)

        :param r_id - the UUID of the object you'd like to update
        :param updated_path - path to the contents of the updated object.

        """
        pass

    def delete(resource_id: str) -> None:
        """
        Delete a copy of a resource from this adapter.
        Delete the corresponding entry in the `copies` table.

        :param r_id - the resource to retrieve's UUID
        """
        pass

    def get_actual_checksum(self, r_id: str) -> str:
        """
        Return an exact checksum of a resource, not relying on the metadata db

        If possible, this should be done with no file I/O
        """
        pass



import logging
from crontab import CronTab
from typing import List

logger = logging.getLogger(__name__)


class Scheduler:
    """
    The LIBREary scheduler is responsible for scheduling various kinds of checks of LIBREarys

    It is meant to be interacted with by users. It currently uses cron jobs to schedule access.

    You must use one scheduler per LIBREary.

    This class currently contains the following methods:

    - set_schedule
    - verify_schedule
    - show_schedule
    """

    def __init__(self, config_dir: str):
        try:
            self.crontab = CronTab(user=True)
        except Exception as e:
            logger.error(
                f"Could not create Libreary Scheduler. Exception: {e}")

    def set_schedule(self, schedule: List[dict]):
        """
        Set a schedule to the scheduler.

        Expects a list of dictionaries, each of which should be structured as follows:
        ```{json}
        {
            "config_dir": "Path to config_directory",
            "levels_to_check": ["list of levels to check"],
            "other_commands":["line here", "line here"],
            "timing": []
        }

        :param schedule - list of dictionaries as described.
        """
        for entry in schedule.items():
            self.add_schedule_job(entry)

    def add_schedule_job(self, schedule_entry: dict):
        """
        Add a single job to the schedule.

        ```{json}
        {
            "config_dir": "Path to config_directory",
            "levels_to_check": ["list of levels to check"],
            "other_commands":["line here", "line here"],
            "timing": [""]
        }
        """
        timing = schedule_entry["timing"]
        command = self.build_single_python_command(schedule_entry)
        job = self.crontab.new(command=command)
        job.setall(" ".join(timing))

    def build_single_python_command(self, schedule_entry: dict):
        """
        Build the python command to go in the user's crontab

        :param schedule_entry - dictionary formatted as described below

        Schedule entry format:

        ```{json}
        {
            "config_dir": "Path to config_directory",
            "levels_to_check": ["list of levels to check"],
            "other_commands":["line here", "line here"],
        }
        ```
        """
        base_python_command = f"python3 -c from libreary import Libreary; l = Libreary('{self.config_dir}');"
        for level in schedule_entry["levels_to_check"]:
            base_python_command += f"l.check_level('{level}');"
        base_python_command += ";".join(schedule_entry["other_commands"])
        return base_python_command



import json
import os
import hashlib
import logging

try:
    import boto3
    from botocore.exceptions import ClientError

except ImportError:
    _boto_enabled = False
else:
    _boto_enabled = True

from libreary.exceptions import ResourceNotIngestedException, ChecksumMismatchException, NoCopyExistsException
from libreary.exceptions import StorageFailedException, ConfigurationError, OptionalModuleMissingException

logger = logging.getLogger(__name__)


class S3Adapter:
    """
        An Adapter allows LIBREary to save copies of digital objects
            to different places across cyberspace. Working with many
            adapters in concert, one should be able do save sufficient
            copies to places they want them.

        S3Adapter allows users to store objects in AWS S3.
    """

    def __init__(self, config: dict, metadata_man: object = None):
        """
        Constructor for S3Adapter. Expects a python dict :param `config`
            in the following format:

        You must have already created the S3 bucket you wish to use for this to work.

        ```{json}
        {
        "metadata": {
            "db_file": "path to metadata db"
        },
        "adapter": {
            "bucket_name": "name of S3 bucket",
            "adapter_identifier": "friendly identifier",
            "adapter_type": "S3Adapter",
            "region": "AWS Region",
            "key_file":"Path to optional AWS key file. See create_session docs for more"
        },
        "options": {
            "dropbox_dir": "path to dropbox directory",
            "output_dir": "path to output directory"
        },
        "canonical":"(boolean) true if this is the canonical adapter"
        }
        """
        self.config = config
        try:
            self.adapter_id = config["adapter"]["adapter_identifier"]
            self.adapter_type = "S3Adapter"
            self.dropbox_dir = config["options"]["dropbox_dir"]
            self.ret_dir = config["options"]["output_dir"]

            if not _boto_enabled:
                raise OptionalModuleMissingException(
                    ['boto3'], "S3 adapter requires the boto3 module.")

            self.profile = self.config["adapter"].get("profile")
            self.key_file = self.config["adapter"].get("key_file")
            self.bucket_name = self.config["adapter"].get("bucket_name")
            self.region = self.config["adapter"].get("region", "us-west-2")

            self.env_specified = os.getenv("AWS_ACCESS_KEY_ID") is not None and os.getenv(
                "AWS_SECRET_ACCESS_KEY") is not None

            self.metadata_man = metadata_man
            if self.metadata_man is None:
                raise KeyError

            logger.debug("Creating S3 Adapter")
        except KeyError:
            logger.error("Invalid configuration for S3 Adapter")
            raise KeyError

        if self.profile is None and self.key_file is None and not self.env_specified:
            raise ConfigurationError("Must specify either profile', 'key_file', or "
                                     "'AWS_ACCESS_KEY_ID' and 'AWS_SECRET_ACCESS_KEY' environment variables.")

        try:
            self.initialize_boto_client()
        except Exception as e:
            logger.error(f"Could not create AWS session")
            raise e
        except ClientError as e:
            logger.error(f"Could not create AWS session - Boto3 client failed.")
            raise e

        self._create_bucket_if_nonexistent()

    def initialize_boto_client(self) -> None:
        """Initialize the boto client."""

        self.session = self.create_session()
        # self.client = self.session.client('s3')
        # self.s3 = self.session.resource('s3')
        self.client = boto3.client('s3')
        self.s3 = boto3.resource('s3')

    def create_session(self) -> boto3.session.Session:
        """Create a session.

        First we look in self.key_file for a path to a json file with the
        credentials. The key file should have 'AWSAccessKeyId' and 'AWSSecretKey'.
        Next we look at self.profile for a profile name and try
        to use the Session call to automatically pick up the keys for the profile from
        the user default keys file ~/.aws/config.
        Finally, boto3 will look for the keys in environment variables:
        AWS_ACCESS_KEY_ID: The access key for your AWS account.
        AWS_SECRET_ACCESS_KEY: The secret key for your AWS account.
        AWS_SESSION_TOKEN: The session key for your AWS account.
        This is only needed when you are using temporary credentials.
        The AWS_SECURITY_TOKEN environment variable can also be used,
        but is only supported for backwards compatibility purposes.
        AWS_SESSION_TOKEN is supported by multiple AWS SDKs besides python.
        """

        session = None

        if self.key_file is not None:
            credfile = os.path.expandvars(os.path.expanduser(self.key_file))

            try:
                with open(credfile, 'r') as f:
                    creds = json.load(f)
            except json.JSONDecodeError as e:
                logger.error(f"Could not create AWS session")
                raise e

            except Exception as e:
                logger.error(f"Could not create AWS session")
                raise e

            session = boto3.session.Session(region_name=self.region, **creds)
        elif self.profile is not None:
            session = boto3.session.Session(
                profile_name=self.profile, region_name=self.region
            )
        else:
            session = boto3.session.Session(region_name=self.region)
        logger.error(f"Created AWS session")
        return session

    def _create_bucket_if_nonexistent(self) -> None:
        """
        Create the S3 bucket we will need if it doesn't already exist
        """
        location = {'LocationConstraint': self.region}
        response = self.client.list_buckets()
        buckets = [bucket['Name'] for bucket in response['Buckets']]

        if self.bucket_name in buckets:
            logger.debug(f"Found existing bucket. ID: {self.bucket_name}")
            return

        logger.debug(f"No existing bucket. Creating: {self.bucket_name}")
        self.s3.create_bucket(Bucket=self.bucket_name,
                              CreateBucketConfiguration=location)

    def store(self, r_id: str) -> None:
        """
        Store a copy of a resource in this adapter.

        Store assumes that the file is in the `dropbox_dir`.
        AdapterManager will always verify that this is the case.

        :param r_id - the resource to store's UUID
        """
        logger.debug(f"Storing object {r_id} to adapter {self.adapter_id}")

        file_metadata = self.metadata_man.get_resource_info(r_id)[0]
        checksum = file_metadata[4]
        name = file_metadata[3]
        current_location = "{}/{}".format(self.dropbox_dir, name)

        sha1Hash = hashlib.sha1(open(current_location, "rb").read())
        sha1Hashed = sha1Hash.hexdigest()

        other_copies = self.metadata_man.get_copy_info(
            r_id, self.adapter_id)
        if len(other_copies) != 0:
            logger.debug(
                f"Other copies of {r_id} from {self.adapter_id} exist")
            return

        if sha1Hashed == checksum:
            locator = '{}_{}'.format(r_id, name)
            self.s3.Bucket(
                self.bucket_name).upload_file(
                current_location,
                locator)
        else:
            logger.error(f"Checksum Mismatch on {r_id} from {self.adapter_id}")
            raise ChecksumMismatchException

        self.metadata_man.add_copy(
            r_id,
            self.adapter_id,
            locator,
            sha1Hashed,
            self.adapter_type,
            canonical=False)

    def _store_canonical(self, current_path: str, r_id: str,
                         checksum: str, filename: str) -> str:
        """
            Store a canonical copy of a resource in this adapter.

            If we're using the S3Adapter as a canonical adapter, we need
            to be able to store from a current path, taking in a generated UUID,
            rather than looking info up from the database.

            :param current_path - current path to object
            :param r_id - UUID of resource you're storing
            :param checksum - checksum of resource
            :param filename - filename of resource you're storing

        """
        logger.debug(
            f"Storing canonical object {r_id} to adapter {self.adapter_id}")
        current_location = current_path

        sha1Hash = hashlib.sha1(open(current_location, "rb").read())
        sha1Hashed = sha1Hash.hexdigest()

        locator = "canonical_{}_{}".format(r_id, filename)

        other_copies = self.metadata_man.get_canonical_copy_metadata(
            r_id)
        if len(other_copies) != 0:
            logger.error(
                f"Other canonical copies of {r_id} from {self.adapter_id} exist")
            raise StorageFailedException

        if sha1Hashed == checksum:
            self.s3.Bucket(self.bucket_name).upload_file(current_path, locator)
        else:
            logger.error(f"Checksum Mismatch on {r_id} from {self.adapter_id}")
            raise ChecksumMismatchException

        self.metadata_man.add_copy(
            r_id,
            self.adapter_id,
            locator,
            sha1Hashed,
            self.adapter_type,
            canonical=True)

        return locator

    def retrieve(self, r_id: str) -> str:
        """
        Retrieve a copy of a resource from this adapter.

        Retrieve assumes that the file can be stored to the `output_dir`.
        AdapterManager will always verify that this is the case.

        Returns the path to the resource.

        May overwrite files in the `output_dir`

        :param r_id - the resource to retrieve's UUID
        """
        try:
            filename = self.metadata_man.get_resource_info(r_id)[0][3]
        except IndexError:
            logger.error(f"Cannot Retrieve object {r_id}. Not ingested.")
            raise ResourceNotIngestedException

        try:
            copy_info = self.metadata_man.get_copy_info(
                r_id, self.adapter_id)[0]
        except IndexError:
            logger.error(
                f"Tried to retrieve a nonexistent copy of {r_id} from {self.adapter_id}")
            raise NoCopyExistsException
        expected_hash = copy_info[4]
        copy_locator = copy_info[3]
        real_hash = copy_info[4]

        new_location = "{}/{}".format(self.ret_dir, filename)

        if real_hash == expected_hash:
            self.s3.Bucket(
                self.bucket_name).download_file(
                copy_locator,
                new_location)
        else:
            logger.error(f"Checksum Mismatch on object {r_id}")
            raise ChecksumMismatchException

        return new_location

    def update(resource_id: str, updated_path: str) -> None:
        """
        Update a resource with a new object. Preserves UUID and all other metadata (levels, etc.)

        :param r_id - the UUID of the object you'd like to update
        :param updated_path - path to the contents of the updated object.

        """
        pass

    def delete(self, r_id: str) -> None:
        """
        Delete a copy of a resource from this adapter.
        Delete the corresponding entry in the `copies` table.

        :param r_id - the resource to retrieve's UUID
        """
        logger.debug(f"Deleting copy of object {r_id} from {self.adapter_id}")

        copy_info = self.metadata_man.get_copy_info(
            r_id, self.adapter_id)

        if len(copy_info) == 0:
            # We've already deleted, probably as part of another level
            return
        locator = copy_info[3]

        self.client.delete_object(
            Bucket=self.bucket_name, Key=locator)
        self.cursor.execute("delete from copies where copy_id=?",
                            [copy_info[0]])
        self.conn.commit()

    def _delete_canonical(self, r_id: str) -> None:
        """
        Delete a canonical copy of a resource from this adapter.
        Delete the corresponding entry in the `copies` table.

        :param r_id - the resource to retrieve's UUID
        """
        logger.debug(
            f"Deleting canonical copy of object {r_id} from {self.adapter_id}")
        try:
            copy_info = self.metadata_man.get_canonical_copy_metadata(
                r_id)[0]
        except IndexError:
            logger.debug(
                f"Canonical copy of {r_id} on {self.adapter_id} has already been deleted.")
            return

        locator = copy_info[3]

        self.client.delete_object(
            Bucket=self.bucket_name, Key=locator)

        self.metadata_man.delete_copy_metadata(copy_info[0])

    def get_actual_checksum(self, r_id: str,
                            delete_after_download: bool = True) -> str:
        """
        Returns an exact checksum of a resource, not relying on the metadata db.

        If possible, implementations of get_actual_checksum should do no file I/O.
            For S3, we need to download and checksum manually. :/

        :param r_id - resource we want the checksum of
        :param delete_after_download - True if the file should be downloaded after the
            checksum is calculated
        """
        logger.debug(
            f"Getting actual checksum of object {r_id} from adapter {self.adapter_id}")
        new_path = self.retrieve(r_id)

        sha1Hash = hashlib.sha1(open(new_path, "rb").read())
        sha1Hashed = sha1Hash.hexdigest()

        if delete_after_download:
            logger.debug(f"Delete after download enabled on {self.adapter_id}")
            os.remove(new_path)

        return sha1Hashed



import os
import hashlib
import pickle
from pathlib import Path
import logging


try:
    from googleapiclient.discovery import build
    from google_auth_oauthlib.flow import InstalledAppFlow
    from google.auth.transport.requests import Request
    from apiclient.http import MediaFileUpload, MediaIoBaseDownload

except ImportError:
    _google_enabled = False
else:
    _google_enabled = True

from libreary.exceptions import ResourceNotIngestedException, ChecksumMismatchException
from libreary.exceptions import StorageFailedException, NoCopyExistsException, OptionalModuleMissingException

# Google Drive Scope
SCOPES = ['https://www.googleapis.com/auth/drive']

logger = logging.getLogger(__name__)


class GoogleDriveAdapter():
    """docstring for GoogleDriveAdapter

        An Adapter allows LIBREary to save copies of digital objects
            to different places across cyberspace. Working with many
            adapters in concert, one should be able do save sufficient
            copies to places they want them.

        DriveAdapter allows you to store objects in Google Drive

    """

    def __init__(self, config: dict, metadata_man: object = None):
        """
        Constructor for GoogleDriveAdapter. Expects a python dict :param `config`
            in the following format:

        You must have already created the Google Drive directory you wish to use for this to work.

        ```{json}
        {
        "metadata": {
            "db_file": "path to metadata db"
        },
        "adapter": {
            "folder_path": "Name of google drive folder for storage. LIBREary will create this folder",
            "adapter_identifier": "friendly identifier",
            "adapter_type": "GoogleDriveAdapter",
            "credentials_file":"Path to credentials file. See get_google_client docs for more",
            "token_file":"Path to place you want to save a token file",
        },
        "options": {
            "dropbox_dir": "path to dropbox directory",
            "output_dir": "path to output directory"
        },
        "canonical":"(boolean) true if this is the canonical adapter"
        }
        """
        try:
            self.adapter_id = config["adapter"]["adapter_identifier"]
            self.token_file = config["adapter"]["token_file"]
            self.dropbox_dir = config["options"]["dropbox_dir"]
            self.folder_path = config["adapter"]["folder_path"]
            self.adapter_type = "GoogleDriveAdapter"
            self.ret_dir = config["options"]["output_dir"]
            self.credentials_file = config["adapter"]["credentials_file"]

            self.metadata_man = metadata_man
            if self.metadata_man is None:
                raise KeyError

            logger.debug("Creating Drive Adapter")
        except KeyError:
            logger.error("Invalid configuration for Drive Adapter")
            raise KeyError

        if not _google_enabled:
            logger.error(
                "Google Drive adapter requires the googleapiclient module.")
            raise OptionalModuleMissingException(
                ['googleapiclient'], "Google Drive adapter requires the googleapiclient module.")

        self.get_google_client()
        self.dir_id = self._get_or_create_folder()

    def get_google_client(self) -> None:
        """
        Build a Google Drive client object.

        Important to note that this uses an OAUTH flow, so you'll
        need to run it from a computer that has a web browser you can use.

        Store the creds JSON file in the place you note in `config["adapter"]["credentials_file"]`
        A token will be stored in `config["adapter"]["token_file"]`.

        If you are running LIBREary on a headless server, I recommend getting a token first,
        and saving the token file on the server, so that you don't need to mess around with
        headless browsers etc.

        Get creds JSON file from here:
            https://developers.google.com/drive/api/v3/quickstart/python?authuser=3
        """
        creds = None
        # The file token.pickle stores the user's access and refresh tokens, and is
        # created automatically when the authorization flow completes for the first
        # time.

        logger.debug("Attempting to acquire google credentials")
        # We save this in the self.config["token_file"] file
        if os.path.exists(self.token_file):
            with open(self.token_file, 'rb') as token:
                creds = pickle.load(token)
        # If there are no (valid) credentials available, let the user log in.
        if not creds or not creds.valid:
            if creds and creds.expired and creds.refresh_token:
                logger.debug("Google credentials expired. Refreshing")
                creds.refresh(Request())
            else:
                logger.debug(
                    "Google credentials not found. Acquiring new token")
                flow = InstalledAppFlow.from_client_secrets_file(
                    self.credentials_file, SCOPES)
                creds = flow.run_local_server(port=0)
            # Save the credentials for the next run
            with open(self.token_file, 'wb') as token:
                logger.debug(f"Saving token to file {self.token_file}")
                pickle.dump(creds, token)

        logger.debug("Building GoogleDrive API service")
        self.service = build('drive', 'v3', credentials=creds)

    def _list_objects(self) -> None:
        """
        Sanity-check method for devs to use. Lists top 1000 items in drive
        """
        logger.debug(f"Testing drive connection")
        results = self.service.files().list(
            pageSize=1000, fields="nextPageToken, files(id, name)").execute()
        items = results.get('files', [])

        if not items:
            logger.info('No files found.')
        else:
            for item in items:
                logger.info(item)

    def _get_or_create_folder(self) -> str:
        """
        If the folder specified in config exists, get its id.
        If not, create it.
        """
        page_token = None
        dir_id = None
        response = self.service.files().list(q="mimeType='application/vnd.google-apps.folder' and name='{}'".format(self.folder_path),
                                             spaces='drive',
                                             fields='nextPageToken, files(id, name)',
                                             pageToken=page_token).execute()
        for file in response.get('files', []):
            dir_id = file.get('id')
            page_token = response.get('nextPageToken', None)
            if page_token is None:
                logger.debug(f"Found existing directory. ID: {dir_id}")
                break

        if not dir_id:

            file_metadata = {
                'name': self.folder_path,
                'mimeType': 'application/vnd.google-apps.folder'
            }
            file = self.service.files().create(body=file_metadata,
                                               fields='id').execute()
            dir_id = file.get('id')
            logger.debug(
                f"Could not find existing directory. 9io9o0Created new one - 9io9o0ID: {dir_id}")

        return dir_id

    def _upload_file(self, filename: str, current_path: str) -> str:
        """
        Helper method to upload a file to drive, in the directory
        LIBRE-ary is configured to use.

        :param filename - name of file to upload
        :param current_path - place where the file is right now
        """
        logger.debug(f"Uploading file {filename} to Drive.")
        file_metadata = {'name': filename,
                         'parents': [self.dir_id]}
        media = MediaFileUpload(current_path,
                                mimetype='image/jpeg')
        file = self.service.files().create(body=file_metadata,
                                           media_body=media,
                                           fields='id').execute()
        f_id = file.get('id')
        return f_id

    def store(self, r_id: str) -> str:
        """
        Store a copy of a resource in this adapter.

        Store assumes that the file is in the `dropbox_dir`.
        AdapterManager will always verify that this is the case.

        :param r_id - the resource to store's UUID
        """
        logger.debug(f"Storing object {r_id} to adapter {self.adapter_id}")

        file_metadata = self.metadata_man.get_resource_info(r_id)[0]
        checksum = file_metadata[4]
        name = file_metadata[3]
        current_location = "{}/{}".format(self.dropbox_dir, name)

        sha1Hash = hashlib.sha1(open(current_location, "rb").read())
        sha1Hashed = sha1Hash.hexdigest()

        new_name = "{}_{}".format(r_id, name)

        other_copies = self.metadata_man.get_copy_info(
            r_id, self.adapter_id)

        if len(other_copies) != 0:
            logger.debug(
                f"Other copies of {r_id} from {self.adapter_id} exist")
            return

        if sha1Hashed == checksum:
            locator = self._upload_file(new_name, current_location)
        else:
            logger.error(f"Checksum Mismatch on {r_id} from {self.adapter_id}")
            raise ChecksumMismatchException

        self.metadata_man.add_copy(
            r_id,
            self.adapter_id,
            locator,
            sha1Hashed,
            self.adapter_type,
            canonical=False)

        return locator

    def retrieve(self, r_id: str) -> str:
        """
        Retrieve a copy of a resource from this adapter.

        Retrieve assumes that the file can be stored to the `output_dir`.
        AdapterManager will always verify that this is the case.

        Returns the path to the resource.

        May overwrite files in the `output_dir`

        :param r_id - the resource to retrieve's UUID
        """
        logger.debug(
            f"Retrieving object {r_id} from adapter {self.adapter_id}")
        try:
            filename = self.metadata_man.get_resource_info(r_id)[0][3]
        except IndexError:
            logger.error(f"Cannot Retrieve object {r_id}. Not ingested.")
            raise ResourceNotIngestedException
        try:
            copy_info = self.metadata_man.get_copy_info(
                r_id, self.adapter_id)[0]
        except IndexError:
            logger.error(
                f"Tried to retrieve a nonexistent copy of {r_id} from {self.adapter_id}")
            raise NoCopyExistsException
        expected_hash = copy_info[4]
        copy_locator = copy_info[3]
        real_hash = copy_info[4]

        new_location = "{}/{}".format(self.ret_dir, filename)

        if real_hash == expected_hash:
            self._download_file(copy_locator, new_location)
        else:
            logger.error(f"Checksum Mismatch on object {r_id}")
            raise ChecksumMismatchException

        return new_location

    def _download_file(self, locator: str, new_loc: str) -> None:
        """
        Helper method to download a file from drive, in the directory
        LIBRE-ary is configured to use.

        :param locator - google drive ID
        :param new_loc - place the file should go
        """
        request = self.service.files().get_media(fileId=locator)
        Path(new_loc).touch()
        fh = open(new_loc, "wb")
        downloader = MediaIoBaseDownload(fh, request)
        done = False
        while done is False:
            status, done = downloader.next_chunk()
            if not status:
                raise ChecksumMismatchException

    def update(self, r_id: str, updated: str) -> None:
        """
        Update a resource with a new object. Preserves UUID and all other metadata (levels, etc.)

        :param r_id - the UUID of the object you'd like to update
        :param updated_path - path to the contents of the updated object.

        """
        pass

    def _store_canonical(self, current_path: str, r_id: str,
                         checksum: str, filename: str) -> str:
        """
            If we're using the GoogleDrive as a canonical adapter, we need
            to be able to store from a current path, taking in a generated UUID,
            rather than looking info up from the database.

            :param current_path - current path to object
            :param r_id - UUID of resource you're storing
            :param checksum - checksum of resource
            :param filename - filename of resource you're storing
        """
        logger.debug(
            f"Storing canonical copy of object {r_id} to {self.adapter_id}")

        new_name = "canonical_{}_{}".format(r_id, filename)

        sha1Hash = hashlib.sha1(open(current_path, "rb").read())
        sha1Hashed = sha1Hash.hexdigest()
        other_copies = self.metadata_man.get_canonical_copy_metadata(
            r_id)
        if len(other_copies) != 0:
            logger.error(
                f"Other canonical copies of {r_id} from {self.adapter_id} exist")
            raise StorageFailedException

        if sha1Hashed == checksum:
            locator = self._upload_file(new_name, current_path)
        else:
            logger.error(f"Checksum Mismatch on object {r_id}")
            raise ChecksumMismatchException

        self.metadata_man.add_copy(
            r_id,
            self.adapter_id,
            locator,
            sha1Hashed,
            self.adapter_type,
            canonical=True)

        return locator

    def delete(self, r_id: str) -> None:
        """
        Delete a copy of a resource from this adapter.
        Delete the corresponding entry in the `copies` table.

        :param r_id - the resource to retrieve's UUID
        """
        logger.debug(f"Deleting copy of object {r_id} from {self.adapter_id}")

        copy_info = self.metadata_man.get_copy_info(
            r_id, self.adapter_id)

        if len(copy_info) == 0:
            # We've already deleted, probably as part of another level
            return

        copy_info = copy_info[0]
        copy_locator = copy_info[3]

        self.service.files().delete(fileId=copy_locator).execute()

        self.metadata_man.delete_copy_metadata(copy_info[0])

    def _delete_canonical(self, r_id: str) -> None:
        """
        Delete a canonical copy of a resource from this adapter.
        Delete the corresponding entry in the `copies` table.

        :param r_id - the resource to retrieve's UUID
        """
        logger.debug(
            f"Deleting canonical copy of object {r_id} from {self.adapter_id}")

        try:
            copy_info = self.metadata_man.get_canonical_copy_metadata(
                r_id)[0]
        except IndexError:
            logger.debug(
                f"Canonical copy of {r_id} on {self.adapter_id} has already been deleted.")
            return

        copy_locator = copy_info[3]

        self.service.files().delete(fileId=copy_locator).execute()

        self.metadata_man.delete_copy_metadata(copy_info[0])

    def get_actual_checksum(self, r_id: str,
                            delete_after_download: bool = True) -> str:
        """
        Return an exact checksum of a resource, not relying on the metadata db.

        The :param deep trusts the tag we've given google drive on ingestion,
        if True, it will retrieve and recompute
        """
        logger.debug(
            f"Getting actual checksum of object {r_id} from adapter {self.adapter_id}")
        new_path = self.retrieve(r_id)

        sha1Hash = hashlib.sha1(open(new_path, "rb").read())
        sha1Hashed = sha1Hash.hexdigest()

        if delete_after_download:
            logger.debug(f"Delete after download enabled on {self.adapter_id}")
            os.remove(new_path)

        return sha1Hashed



from typing import List


class BaseMetadataManager(object):
    """docstring for BaseMetadataManager

    This object contains the following methods:

    - verify_db_structure

    """

    def __init__(self, config: dict):
        pass

    def verify_db_structure(self) -> bool:
        pass

    def add_level(self, name: str, frequency: int,
                  adapters: List[dict], copies=1) -> None:
        pass

    def ingest_to_db(self, canonical_adapter_locator: str,
                     levels: List[str], filename: str, checksum: str, obj_uuid: str, description: str) -> None:
        pass

    def list_resources(self) -> List[List[str]]:
        pass

    def get_resource_info(self, r_id: str) -> List[str]:
        pass

    def delete_resource(self, r_id: str) -> None:
        pass

    def minimal_test_ingest(self, locator: str, real_checksum: str, r_id: str):
        pass

    def get_levels(self):
        pass

    def update_resource_levels(self, r_id: str, new_levels: List[str]):
        pass

    def summarize_copies(self, r_id: str) -> List[List[str]]:
        pass

    def get_canonical_copy_metadata(self, r_id: str) -> List[List[str]]:
        pass

    def get_copy_info(self, r_id: str, adapter_id: str,
                      canonical: bool = False):
        pass

    def delete_copy_metadata(self, copy_id: str):
        pass

    def add_copy(self, r_id: str, adapter_id: str, new_location: str,
                 sha1Hashed: str, adapter_type: str, canonical: bool = False):
        pass

    def search(self, search_term: str) -> List[List[str]]:
        pass



class IngestionFailedException(Exception):
    pass


class ChecksumMismatchException(Exception):
    pass


class MetadataUnavailableException(Exception):
    pass


class AdapterConnectionFailureException(Exception):
    pass


class RestorationFailedException(Exception):
    pass


class AdapterCreationFailedException(Exception):
    pass


class StorageFailedException(Exception):
    pass


class OptionalModuleMissingException(Exception):
    pass


class ResourceNotIngestedException(Exception):
    pass


class NoCopyExistsException(Exception):
    pass


class AdapterRestored(Exception):
    pass


class ConfigurationError(Exception):
    pass


class NoSuchMetadataFieldExeption(Exception):
    pass



import sqlite3
import os
import json
from typing import List
import logging

from libreary.exceptions import ResourceNotIngestedException, NoSuchMetadataFieldExeption

logger = logging.getLogger(__name__)


class SQLite3MetadataManager(object):
    """docstring for SQLite3MetadataManager

    SQLite3 Metadata Manager is the most basic.

    It expects a SQLite3 file, formatted as described in
    the LIBREary documentation.
    """

    def __init__(self, config: dict):
        """
        Constructor for the MedadataManager object. This object can be created manually, but
        in most cases, it will be constructed by the LIBRE-ary main object. It expects a python dict
        :param config, which should be structured as follows:
        ```{json}
        {
        "db_file": "path to SQLite3 DB file for metadata"
        }
        ```
        """
        try:
            self.metadata_db = os.path.realpath(
                config.get("db_file"))
            self.conn = sqlite3.connect(self.metadata_db)
            self.cursor = self.conn.cursor()
            self.type = config.get("manager_type")
            logger.debug(
                "Metadata Manager Configuration Valid. Creating Metadata Manager")
        except KeyError:
            logger.error("Ingester Configuration Invalid")
            raise KeyError

    def verify_db_structure(self) -> bool:
        pass

    def add_level(self, name: str, frequency: int,
                  adapters: List[dict], copies=1) -> None:
        """
        Add a level to the metadata database.

        :param name - name for the level
        :param frequency - check frequency for level. Currently unimplemented
        :param adapters - dict object specifying adapters the level uses. Example:
            ```{json}
            [
                {
                "id": "local1",
                "type":"LocalAdapter"
                },
                {
                "id": "local2",
                "type":"LocalAdapter"
                }
            ]

            ```
        :param copies - copies to store for each adapter. Currently, only 1 is supported
        """
        logger.debug(f"Adding level {name}")
        str_adapters = json.dumps(adapters)
        self.cursor.execute(
            "insert into levels values (?, ?, ?, ?, ?)",
            (None,
             name,
             frequency,
             str_adapters,
             copies))
        self.conn.commit()

    def delete_level(self, name: str) -> None:
        """
        Delete a level from the metadata database.

        Deletes level, also deletes level from levels in resources

        :param name - name of level to delete
        """
        logger.debug(f"Deleting level {name}")
        self.cursor.execute(
            "delete from levels where name=?",
            (name,))
        self.conn.commit()
        resources = self.list_resources()

        for resource in resources:
            uuid = resource[5]
            try:
                levels = resource[2].split(",")
                levels.remove(name)
            except IndexError:
                continue
            except ValueError:
                continue
            print(levels)
            self.cursor.execute(
                "update resources set levels=? where uuid=?", (json.dumps(levels), str(uuid)))

    def ingest_to_db(self, canonical_adapter_locator: str,
                     levels: str, filename: str, checksum: str, obj_uuid: str, description: str) -> None:
        """
        Ingest an object's metadata to the metadata database.

        :param canonical_adapter_locator - locator from the canonical adapter.
               usually something like a file path or object ID
        :param levels - list of strings, each the name of a level the object should be stored at
        :param filename - the filename of the object to be ingested
        :param checksum - the checksum of object to be ingested
        :param UUID - the UUID to be tagged with the object
        :param description - a friendly, searchable description of the object
        """
        logger.debug(f"Ingesting object {obj_uuid} with name {filename}")
        self.cursor.execute("insert into resources values (?, ?, ?, ?, ?, ?, ?)",
                            (None, canonical_adapter_locator, levels, filename, checksum, obj_uuid, description))

        self.conn.commit()

    def list_resources(self) -> List[List[str]]:
        """
        Return a list of summaries of each resource. This summary includes:

        `id`, `path`, `levels`, `file name`, `checksum`, `object uuid`, `description`

        This method trusts the metadata database. There should be a separate method to
        verify the metadata db so that we know we can trust this info
        """
        return self.cursor.execute("select * from resources").fetchall()

    def get_resource_info(self, r_id: str) -> List[str]:
        """
        Get all of the resource metadata for a resource That summary includes:

        `id`, `path`, `levels`, `file name`, `checksum`, `object uuid`, `description`

        This method trusts the metadata database. There should be a separate method to
        verify the metadata db so that we know we can trust this info

        This returns metadata that's kept in the `resources` table, not the `copies` table
        """
        return self.cursor.execute(
            "select * from resources where uuid=?", (r_id,)).fetchall()

    def delete_resource(self, r_id: str) -> None:
        """
        Delete a resource's metadata from the `resources` table

        :param r_id - the resource's uuid
        """
        self.cursor.execute("delete from resources where uuid=?", (r_id,))
        self.conn.commit()

    def minimal_test_ingest(self, locator: str, real_checksum: str, r_id: str):
        """
        Minimally ingest a resource for adapter testing

        Adapters depend on certain information in the `resources` table.

        In order for adapter manager to test adapters, it needs to have minimally ingested them.
        This testing has predefined levels, filenames, and descriptions.

        In general, I don't advise running this method yourself. Allow the adapter manager to do it.

        :param locator - locator of the copy you're testing
        :param real_checksum - object checksum
        :param r_id - r_id of test resource
        """
        self.cursor.execute("insert into resources values (?, ?, ?, ?, ?, ?, ?)",
                            (None, locator, "low,", "libreary_test_file.txt", real_checksum, r_id, "A resource for testing LIBREary adapters with"))
        self.conn.commit()

    def get_levels(self):
        """
        Return all configured levels
        """
        return self.cursor.execute("select * from levels").fetchall()

    def update_resource_levels(self, r_id: str, new_levels: List[str]):
        """
        Change a resource's levels

        :param r_id - resource id of resource to update
        :param new_levels - list of names of new levels
        """
        sql = "update resources set levels = ? where uuid=?"
        self.cursor.execute(sql, (r_id, ",".join([l for l in new_levels])))
        self.conn.commit()

    def summarize_copies(self, r_id: str) -> List[List[str]]:
        """
        Get a summary of all copies of a single resource. That summary includes:

        `copy_id`, `resource_id`, `adapter_identifier`, `locator`, `checksum`, `adapter type`, `canonical (bool)`
        for each copy

        This method trusts the metadata database. There should be a separate method to
        verify the metadata db so that we know we can trust this info

        :param r_id - UUID of resource you'd like to learn about
        """
        sql = "select * from copies where resource_id = ?"
        return self.cursor.execute(sql, (r_id,)).fetchall()

    def get_canonical_copy_metadata(self, r_id: str) -> List[List[str]]:
        """
        Get a summary of the canonical copy of an object's medatada. That summary includes:
        `copy_id`, `resource_id`, `adapter_identifier`, `locator`, `checksum`, `adapter type`, `canonical (bool)`

        :param r_id - UUID of resource you'd like to learn about
        """
        sql = "select * from copies where resource_id = ? and canonical=1"
        return self.cursor.execute(sql, (
            r_id,)).fetchall()

    def get_copy_info(self, r_id: str, adapter_id: str):
        """
        Get a summary of a copy of an object. Can be canonical or not.

        :param r_id - object you want to learn about
        :param adapter_id - adapter storing the copy
        """
        return self.cursor.execute(
            "select * from copies where resource_id=? and adapter_identifier=?", [
                r_id, adapter_id]).fetchall()

    def delete_copy_metadata(self, copy_id: int):
        """
        Delete object metadata for a single copy

        :param copy_id -  The copy id (not resource uuid) to delete
        """
        print(copy_id)
        self.cursor.execute("delete from copies where copy_id=?",
                            (copy_id,))
        self.conn.commit()

    def add_copy(self, r_id: str, adapter_id: str, new_location: str,
                 sha1Hashed: str, adapter_type: str, canonical: bool = False):
        """
        Add a copy of an object to the metadata database


        """
        self.cursor.execute(
            "insert into copies values ( ?, ?, ?, ?, ?, ?, ?)",
            [None, r_id, adapter_id, new_location, sha1Hashed, adapter_type, canonical])
        self.conn.commit()

    def search(self, search_term: str):
        """
        Search the metadata db for information about resources.

        :param search_term - a string with which to search against the metadata db.
            Can match UUID, filename, original path, or description.
        """
        search_term = "%" + search_term + "%"
        return self.cursor.execute(f"select * from resources where name like ? or path like ? or uuid like ? or description like ?",
                                   (search_term, search_term, search_term, search_term)).fetchall()

    def list_object_metadata_schema(self, r_id: str) -> List:
        """
        Get a list of all of the object metadata fields related to the object.

        These fields live in the "object_metadata_schema" table, and the related
            data lives in the "object_metadata" table

        :param r_id - object uuid for looking up
        """
        try:
            text_fields = self.cursor.execute(
                f"select md_schema from object_metadata_schema where object_id=?", (r_id,)).fetchall()[0][0]
            return json.loads(text_fields)
        except IndexError:
            return []

    def set_object_metadata_schema(self, r_id: str, md_schema: str) -> None:
        """
        insert a list of all of the object metadata fields related to the object.

        These fields live in the "object_metadata_schema" table, and the related
            data lives in the "object_metadata" table

        :param r_id - object uuid for looking up
        :param md_schema - list of field names and types
        """
        md_schema_text = json.dumps(md_schema)
        self.cursor.execute(
            "insert into object_metadata_schema values ( ?, ?, ?)",
            [None, r_id, md_schema_text])
        self.conn.commit()

    def set_object_metadata_field(
            self, r_id: str, field: str, value: str) -> None:
        """
        Set a single metadata field.

        These fields live in the "object_metadata_schema" table, and the related
            data lives in the "object_metadata" table

        :param r_id - object uuid for looking up
        :param md_schema - list of field names and types
        """
        try:
            list_of_fields = self.list_object_metadata_schema(r_id)
        except Exception:
            raise ResourceNotIngestedException

        if field not in list_of_fields:
            raise NoSuchMetadataFieldExeption

        self.cursor.execute(
            "insert into object_metadata values ( ?, ?, ?, ?)",
            [None, r_id, field, value])
        self.conn.commit()

    def set_all_object_metadata(
            self, r_id: str, metadata: List[dict]) -> None:
        """
        Set all metadata fields.

        These fields live in the "object_metadata_schema" table, and the related
            data lives in the "object_metadata" table

        :param r_id - object uuid for looking up
        :param md_schema - list of field names and types
        """
        try:
            self.list_object_metadata_schema(r_id)
        except Exception:
            raise ResourceNotIngestedException

        for field in metadata:
            self.set_object_metadata_field(
                r_id, field["field"], field["value"])

    def delete_object_metadata(self, r_id: str) -> None:
        """
        Delete object metadata - remove all fields from the "object_metadata" table

        :param r_id - orbect uuid for deletion
        """
        try:
            list_of_fields = self.list_object_metadata_schema(r_id)
        except Exception:
            raise ResourceNotIngestedException
        for field in list_of_fields:
            self.delete_object_metadata_field(r_id, field)

    def delete_object_metadata_field(self, r_id: str, field: str) -> None:
        """
        Delete object metadata - remove single fields from the "object_metadata" table

        :param r_id - orbect uuid for deletion
        :param field - field to delete
        """
        self.cursor.execute(
            "delete from object_metadata where object_id=? and key=?", (r_id, field))
        self.conn.commit()

    def delete_object_metadata_schema(self, r_id: str):
        """
        Delete object metadata schema - remove entry from the "object_metadata_schema" table

        :param r_id - orbect uuid for deletion
        """
        self.cursor.execute(
            "delete from object_metadata_schema where object_id=?", (r_id,))
        self.conn.commit()

    def delete_object_metadata_entirely(self, r_id: str) -> None:
        """
        Delete object metadata schema and all associated metadata

        :param r_id - orbect uuid for deletion
        """
        self.delete_object_metadata(r_id)
        self.delete_object_metadata_schema(r_id)



import logging
from typing import Optional

from libreary.version import VERSION
from libreary.libreary import Libreary
from libreary.exceptions import *
from libreary.adapter_manager import AdapterManager
from libreary.ingester import Ingester
from libreary.scheduler import Scheduler

__author__ = 'Ben Glick'
__version__ = VERSION


AUTO_LOGNAME = -1

"""
Logging
-------
Following the general logging philosophy of python libraries, by default
LIBREary doesn't log anything.

However the following helper functions are provided for logging:
1. set_stream_logger
    This sets the logger to the StreamHandler. This is quite useful when working from
    a Jupyter notebook.
2. set_file_logger
    This sets the logging to a file. This is ideal for reporting issues to the dev team.
Constants
---------
AUTO_LOGNAME
    Special value that indicates libreary should construct a filename for logging.
"""


def set_stream_logger(name: str = 'libreary', level: int = logging.DEBUG,
                      format_string: Optional[str] = None):
    """Add a stream log handler.
    Args:
         - name (string) : Set the logger name.
         - level (logging.LEVEL) : Set to logging.DEBUG by default.
         - format_string (string) : Set to None by default.
    Returns:
         - None
    """
    if format_string is None:
        # format_string = "%(asctime)s %(name)s [%(levelname)s] Thread:%(thread)d %(message)s"
        format_string = "%(asctime)s %(name)s:%(lineno)d [%(levelname)s]  %(message)s"

    logger = logging.getLogger(name)
    logger.setLevel(logging.DEBUG)
    handler = logging.StreamHandler()
    handler.setLevel(level)
    formatter = logging.Formatter(format_string, datefmt='%Y-%m-%d %H:%M:%S')
    handler.setFormatter(formatter)
    logger.addHandler(handler)

    # Concurrent.futures errors are also of interest, as exceptions
    # which propagate out of the top of a callback are logged this way
    # and then discarded. (see #240)
    futures_logger = logging.getLogger("concurrent.futures")
    futures_logger.addHandler(handler)


def set_file_logger(filename: str, name: str = 'libreary',
                    level: int = logging.DEBUG, format_string: Optional[str] = None):
    """Add a stream log handler.
    Args:
        - filename (string): Name of the file to write logs to
        - name (string): Logger name
        - level (logging.LEVEL): Set the logging level.
        - format_string (string): Set the format string
    Returns:
       -  None
    """
    if format_string is None:
        format_string = "%(asctime)s.%(msecs)03d %(name)s:%(lineno)d [%(levelname)s]  %(message)s"

    logger = logging.getLogger(name)
    logger.setLevel(logging.DEBUG)
    handler = logging.FileHandler(filename)
    handler.setLevel(level)
    formatter = logging.Formatter(format_string, datefmt='%Y-%m-%d %H:%M:%S')
    handler.setFormatter(formatter)
    logger.addHandler(handler)

    # see note in set_stream_logger for notes about logging
    # concurrent.futures
    futures_logger = logging.getLogger("concurrent.futures")
    futures_logger.addHandler(handler)


class NullHandler(logging.Handler):
    """Setup default logging to /dev/null since this is library."""

    def emit(self, record):
        pass


__all__ = ["Libreary", "AdapterManager", "Ingester", "Scheduler", "set_stream_logger",
           "set_file_logger", "AUTO_LOGNAME"]


logging.getLogger('libreary').addHandler(NullHandler())



from setuptools import setup, find_packages

with open('libreary/version.py') as f:
    exec(f.read())

with open('requirements.txt') as f:
    install_requires = f.readlines()

setup(
    name='libreary',
    version=VERSION,
    description='Distributed Digital Object Archive System',
    long_description='Distributed Digital Object Archive System',
    url='https://github.com/LIBREary/libre-ary',
    author='Ben Glick',
    install_requires=install_requires,
    author_email='glick@glick.cloud',
    license='Apache 2.0',
    download_url='https://github.com/LIBREary/libre-ary#{}'.format(VERSION),
    include_package_data=True,
    packages=find_packages(),
    classifiers=[
        # Maturity
        'Development Status :: 3 - Alpha',
        # Intended audience
        'Intended Audience :: Developers',
        # Licence, must match with licence above
        'License :: OSI Approved :: Apache Software License',
        # Python versions supported
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    keywords=['Archive', 'Distributed Systems']
)



import os
from shutil import copyfile
import hashlib
import logging

from libreary.exceptions import ResourceNotIngestedException, ChecksumMismatchException
from libreary.exceptions import StorageFailedException, NoCopyExistsException

logger = logging.getLogger(__name__)


class LocalAdapter:
    """
        An Adapter allows LIBREary to save copies of digital objects
            to different places across cyberspace. Working with many
            adapters in concert, one should be able do save sufficient
            copies to places they want them.

        LocalAdapter is a basic adapter which saves files
        to a local directory specified in the adapter's config

        Later in this project's plan, the LocalAdapter will be used
        for ingesting the master copies as well as as a (probably)
        commonly used adapter.

        It's also very nice to use for testing, as saving files is easy (ish)
        to debug and doesn't cost any money (unlike a cloud service) or have
        any configuration difficulty.
    """

    def __init__(self, config: dict, metadata_man: object = None):
        """
        Constructor for LocalAdapter. Expects a python dict :param `config`
            in the following format:

        ```{json}
        {
        "metadata": {
            "db_file": "path to metadata db"
        },
        "adapter": {
            "storage_dir": "Directory to store objects in",
            "adapter_identifier": "Friendly identifier",
            "adapter_type": "LocalAdapter"
        },
        "options": {
            "dropbox_dir": "path to dropbox directory",
            "output_dir": "path to output directory"
        },
        "canonical":"(boolean) true if this is the canonical adapter"
        }
        ```

        """
        try:
            self.adapter_id = config["adapter"]["adapter_identifier"]
            self.storage_dir = config["adapter"]["storage_dir"]
            self.dropbox_dir = config["options"]["dropbox_dir"]
            self.adapter_type = "LocalAdapter"
            self.ret_dir = config["options"]["output_dir"]

            self.metadata_man = metadata_man
            if self.metadata_man is None:
                raise KeyError

            logger.debug("Creating Local Adapter")
        except KeyError:
            logger.error("Invalid configuration for Local Adapter")
            raise KeyError

    def store(self, r_id: str) -> str:
        """
        Store a copy of a resource in this adapter.

        Store assumes that the file is in the `dropbox_dir`.
        AdapterManager will always verify that this is the case.

        :param r_id - the resource to store's UUID
        """
        logger.debug(f"Storing object {r_id} to adapter {self.adapter_id}")
        file_metadata = self.metadata_man.get_resource_info(r_id)[0]
        checksum = file_metadata[4]
        name = file_metadata[3]
        current_location = "{}/{}".format(self.dropbox_dir, name)
        new_location = os.path.expanduser(
            "{}/{}".format(self.storage_dir, name))
        new_dir = os.path.expanduser("/".join(new_location.split("/")[:-1]))

        sha1Hash = hashlib.sha1(open(current_location, "rb").read())
        sha1Hashed = sha1Hash.hexdigest()

        if not os.path.isdir(new_dir):
            os.makedirs(new_dir)

        if os.path.isfile(new_location):
            new_location = "{}_{}_{}".format("/".join(new_location.split('/')[:-1]), name, r_id)

        other_copies = self.metadata_man.get_copy_info(
            r_id, self.adapter_id)

        if len(other_copies) != 0:
            logger.debug(
                f"Other copies of {r_id} from {self.adapter_id} exist")
            return

        if sha1Hashed == checksum:
            copyfile(current_location, new_location)
        else:
            logger.error(f"Checksum Mismatch on object {r_id}")
            raise ChecksumMismatchException

        self.metadata_man.add_copy(
            r_id,
            self.adapter_id,
            new_location,
            sha1Hashed,
            self.adapter_type,
            canonical=False)

    def retrieve(self, r_id: str) -> str:
        """
        Retrieve a copy of a resource from this adapter.

        Retrieve assumes that the file can be stored to the `output_dir`.
        AdapterManager will always verify that this is the case.

        Returns the path to the resource.

        May overwrite files in the `output_dir`

        :param r_id - the resource to retrieve's UUID
        """
        logger.debug(
            f"Retrieving object {r_id} from adapter {self.adapter_id}")
        try:
            filename = self.metadata_man.get_resource_info(r_id)[0][3]
        except IndexError:
            logger.error(f"Cannot Retrieve object {r_id}. Not ingested.")
            raise ResourceNotIngestedException
        try:
            copy_info = self.metadata_man.get_copy_info(
                r_id, self.adapter_id)[0]
        except IndexError:
            logger.error(
                f"Tried to retrieve a nonexistent copy of {r_id} from {self.adapter_id}")
            raise NoCopyExistsException
        expected_hash = copy_info[4]
        copy_path = copy_info[3]
        real_hash = copy_info[4]

        new_location = "{}/{}".format(self.ret_dir, filename)

        if real_hash == expected_hash:
            copyfile(copy_path, new_location)
        else:
            logger.error(f"Checksum Mismatch on object {r_id}")
            raise ChecksumMismatchException

        return new_location

    def update(self, r_id: str, updated_path: str) -> None:
        """
        Update a resource with a new object. Preserves UUID and all other metadata (levels, etc.)

        :param r_id - the UUID of the object you'd like to update
        :param updated_path - path to the contents of the updated object.

        """
        pass

    def _store_canonical(self, current_path: str, r_id: str,
                         checksum: str, filename: str) -> str:
        """
            Store a canonical copy of a resource in this adapter.

            If we're using the LocalAdapter as a canonical adapter, we need
            to be able to store from a current path, taking in a generated UUID,
            rather than looking info up from the database.

            :param current_path - current path to object
            :param r_id - UUID of resource you're storing
            :param checksum - checksum of resource
            :param filename - filename of resource you're storing

        """
        logger.debug(
            f"Storing canonical copy of object {r_id} to {self.adapter_id}")
        current_location = current_path
        new_location = os.path.expanduser(
            "{}/canonical_{}".format(self.storage_dir, filename))
        new_dir = os.path.expanduser("/".join(new_location.split("/")[:-1]))

        sha1Hash = hashlib.sha1(open(current_location, "rb").read())
        sha1Hashed = sha1Hash.hexdigest()

        if not os.path.isdir(new_dir):
            os.makedirs(new_dir)

        if os.path.isfile(new_location):
            new_location = "{}_{}".format(new_location, r_id)

        other_copies = self.metadata_man.get_canonical_copy_metadata(
            r_id)
        if len(other_copies) != 0:
            logger.error(
                f"Other canonical copies of {r_id} from {self.adapter_id} exist")
            raise StorageFailedException

        if sha1Hashed == checksum:
            copyfile(current_location, new_location)
        else:
            logger.error(f"Checksum Mismatch on object {r_id}")
            raise ChecksumMismatchException

        self.metadata_man.add_copy(
            r_id,
            self.adapter_id,
            new_location,
            sha1Hashed,
            self.adapter_type,
            canonical=True)

        return new_location

    def delete(self, r_id: str) -> None:
        """
        Delete a copy of a resource from this adapter.
        Delete the corresponding entry in the `copies` table.

        :param r_id - the resource to retrieve's UUID
        """
        logger.debug(f"Deleting copy of object {r_id} from {self.adapter_id}")
        copy_info = self.metadata_man.get_copy_info(
            r_id, self.adapter_id)

        if len(copy_info) == 0:
            # We've already deleted, probably as part of another level
            return

        copy_info = copy_info[0]

        copy_path = copy_info[3]

        os.remove(copy_path)

        self.metadata_man.delete_copy_metadata(copy_info[0])

    def _delete_canonical(self, r_id: str) -> None:
        """
        Delete a canonical copy of a resource from this adapter.
        Delete the corresponding entry in the `copies` table.

        :param r_id - the resource to retrieve's UUID
        """
        logger.debug(
            f"Deleting canonical copy of object {r_id} from {self.adapter_id}")
        copy_info = self.metadata_man.get_canonical_copy_metadata(
            r_id)[0]
        copy_path = copy_info[3]

        os.remove(copy_path)

        self.metadata_man.delete_copy_metadata(copy_info[0])

    def get_actual_checksum(self, r_id: str) -> str:
        """
        Returns an exact checksum of a resource, not relying on the metadata db.

        If possible, implementations of get_actual_checksum should do no file I/O.
            In the case of LocalAdapter, we're able to do this without copying files
            around.

        :param r_id - resource we want the checksum of
        """
        logger.debug(
            f"Getting actual checksum of object {r_id} from adapter {self.adapter_id}")
        copy_info = self.metadata_man.get_copy_info(r_id, self.adapter_id)
        if len(copy_info) == 0:
            raise NoCopyExistsException
        copy_info = copy_info[0]
        path = copy_info[3]
        hash_obj = hashlib.sha1(open(path, "rb").read())
        checksum = hash_obj.hexdigest()
        return checksum



from libreary.adapters.local import LocalAdapter
from libreary.adapters.s3 import S3Adapter
from libreary.adapters.drive import GoogleDriveAdapter


__all__ = ['S3Adapter',
           'LocalAdapter',
           'GoogleDriveAdapter']
